{% extends "!layout.html" %}

{%- block extrahead %}
{{ super() }}

<script type="text/javascript">

  var _gaq = _gaq || [];
  _gaq.push(['_setAccount', 'UA-32855587-1']);
  _gaq.push(['_setDomainName', 'pycsw.org']);
  _gaq.push(['_trackPageview']);

  (function() {
    var ga = document.createElement('script'); ga.type = 'text/javascript'; ga.async = true;
    ga.src = ('https:' == document.location.protocol ? 'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
    var s = document.getElementsByTagName('script')[0]; s.parentNode.insertBefore(ga, s);
  })();

</script>

{% endblock %}

{% block relbar1 %}

<style type="text/css">
  .header a:link{color: white;}
  .header a:visited{color: white;}
  .padded{padding: 10px 10px;}
</style>

{{ super() }}
{% endblock %}

{% block footer %}

<div class="footer">
    <p>&copy; Copyright {{ copyright }}<br/>Last updated on {{ last_updated }}</p>
    <p>
        <a href="https://pycsw.org">Website</a> &mdash;
        <a href="https://pycsw.org/community">Community</a> &mdash;
        <a href="https://pycsw.org/blog">Blog</a> &mdash;
        <a href="https://plus.google.com/communities/104084873011085696113">Google+</a> &mdash;
        <a href="https://pycsw.org/download">Download</a> &mdash;
        <a href="https://github.com/geopython/pycsw">Source</a> &mdash;
        <a href="https://github.com/geopython/pycsw/issues">Issues</a> &mdash;
        <a href="http://lists.osgeo.org/mailman/listinfo/pycsw-devel">Mailing List</a>
    </p>
</div>
{% endblock %}
